import json

from bs4 import BeautifulSoup

from zdict.dictionary import DictBase
from zdict.exceptions import NotFoundError
from zdict.models import Record


# [TODO]
#
# * let user choose en <-> spanish
# * some word's webpage use different CSS class ... (e.g. yo)
# * make code much more readable


class SpanishDict(DictBase):
    '''
    Tested words : ('soy', 'manzana', 'python', 'perdón')
    '''

    HOMEPAGE_URL = "https://www.spanishdict.com/"
    API = 'https://www.spanishdict.com/translate/{word}'

    @property
    def provider(self):
        return 'spanish'

    @property
    def title(self):
        return 'SpanishDict'

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)
        self.color.print(content['word'], 'yellow')

        explains = content.get('explains')
        for data in explains:

            self.color.print(data[0], 'lwhite')     # speech

            for explain in data[1]:
                self.color.print(explain[0], 'lred')    # category
                for t in explain[1]:
                    # t = (index, (sentence, sentence, ...))
                    self.color.print(  # index
                        '{text}'.format(text=t[0]),
                        'org',
                        indent=2
                    )
                    if len(t) > 1:
                        for sentence in t[1]:
                            # sentence = (spanish, english)
                            self.color.print(
                                '{text}'.format(text=sentence[0]),
                                'lindigo',
                                indent=4
                            )
                            self.color.print(
                                '{text}'.format(text=sentence[1]),
                                'indigo',
                                indent=4
                            )
        print()

    def query(self, word: str):
        webpage = self._get_raw(word)
        soup = BeautifulSoup(webpage, "html.parser")
        content = {}

        en_css = "#dictionary-neodict-en"
        es_css = "#dictionary-neodict-es"
        card = soup.select_one(en_css) or soup.select_one(es_css)
        if card is None:
            raise NotFoundError(word)

        word_css = "div > div:nth-child(1) > span"
        word_element = card.select_one(word_css)
        if word_element is None:
            raise NotFoundError(word)
        content['word'] = word_element.text

        '''
        COPULAR VERB  # speech
            # categories_card
            1. (used to express a permanent quality)  # category_text
                # explanation
                a. ser  # index
                # examples
                    # example
                    The ocean is blue.
                    El océano es azul.
            2. (used to express a temporary state)
                a. estar
                    I'm not in a good mood today.
                    Hoy no estoy de buen humor.

                    The sky is cloudy.
                    El cielo está nublado.
        ... (Another speech if it has.)
        '''
        speech_pattern = "div > div:nth-child(2)"
        # "#dictionary-neodict-en > div > div:nth-child(2)"

        # Start to grab
        content['explains'] = []
        speech = card.select_one(speech_pattern)
        while speech:
            result = []
            speech_text, categories_card = speech.children
            speech_text_element = speech_text.find(['a', 'span'])
            content['explains'].append([speech_text_element.text, result])

            for category in categories_card.children:
                category_text_element, explanations_card = category.children
                category_text = category_text_element.text

                explains = []
                for explanation in explanations_card.children:
                    for _ in explanation.children:
                        index_elements, examples = (
                            _.contents[:-1], _.contents[-1]
                        )
                        index = ' '.join(
                            [
                                _.text.strip() for _ in index_elements
                                if _ != ' '
                            ]
                        )

                        if (not examples) and index:
                            explains.append((index,))
                            continue

                        sentences = []
                        for example in examples:
                            t = example.find_all()
                            # Should be only 3 elements
                            # [text, —,  text]
                            '''
                            When Spanish => English, it will show Spanish first
                            When English => Spanish, it will show English first
                            So, the variables below are not definitely
                            '''
                            sentences.append((t[0].text, t[2].text))
                        explains.append((index, sentences))

                result.append([category_text, explains])
            speech = speech.next_sibling

        record = Record(
            word=word,
            content=json.dumps(content),
            source=self.provider,
        )

        return record



import json

from zdict.dictionary import DictBase
from zdict.exceptions import NotFoundError
from zdict.models import Record


class UrbanDict(DictBase):

    HOMEPAGE_URL = "https://www.urbandictionary.com/"
    API = 'http://api.urbandictionary.com/v0/define?term={word}'

    @property
    def provider(self):
        return 'urban'

    @property
    def title(self):
        return 'Urban Dictionary'

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        data = content['list'][0]

        # print word
        self.color.print(data.get('word', ''), 'yellow')

        self.color.print(
            data.get('definition', ''),
            'org',
            indent=2,
        )

        for example in data.get('example', '').split('\n'):
            self.color.print(
                example,
                'indigo',
                indent=2,
            )

        print()

    def query(self, word: str):
        content_str = self._get_raw(word)
        content_dict = json.loads(content_str)

        if content_dict['list'] == []:
            raise NotFoundError(word)

        record = Record(
            word=word,
            content=content_str,
            source=self.provider,
        )

        return record



from pytest import raises
from unittest.mock import patch

from zdict.exceptions import NotFoundError
from zdict.zdict import get_args

from zdict.dictionaries.naer import NaerDict


class TestNaerDict:
    @classmethod
    def setup_class(cls):
        cls.dict = NaerDict(get_args())

        # You may want to change words to some certain test cases.
        cls.words = ['西爾河', 'spring mass']
        cls.not_found_word = 'dsafwwe'

        # Setup normal query data
        cls.dict.args.verbose = False
        cls.records = [cls.dict.query(word) for word in cls.words]

        # Setup verbose query data
        cls.dict.args.verbose = True
        cls.verbose_records = [cls.dict.query(word) for word in cls.words]

        # Change back to default verbose config
        cls.dict.args.verbose = False

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.words
        del cls.not_found_word
        del cls.records
        del cls.verbose_records

    def test_provider(self):
        assert self.dict.provider == 'naer'

    def test_title(self):
        assert self.dict.title == '國家教育研究院 - 雙語詞彙、學術名詞暨辭書資訊網'

    def test__get_url(self):
        url = 'https://terms.naer.edu.tw/search?q=test&field=ti&match=smart'
        assert url == self.dict._get_url('test')

    def test_show(self):
        self.dict.args.verbose = False

        for record in self.records:
            self.dict.show(record)

    def test_show_verbose(self):
        self.dict.args.verbose = True

        for record in self.verbose_records:
            self.dict.show(record)

    @patch('zdict.dictionaries.naer.Record')
    def test_query_normal(self, Record):
        self.dict.args.verbose = False

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word,
                content=self.records[i].content,
                source='naer',
            )

    @patch('zdict.dictionaries.naer.Record')
    def test_query_verbose(self, Record):
        self.dict.args.verbose = True

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word,
                content=self.verbose_records[i].content,
                source='naer',
            )

    def test_query_not_found(self):
        # Trigger NotFoundError intentionally and see if it works.
        with raises(NotFoundError):
            self.dict.query(self.not_found_word)



import json
import requests
from collections import defaultdict

from bs4 import BeautifulSoup

from zdict.dictionary import DictBase
from zdict.exceptions import NotFoundError
from zdict.models import Record


class NaerDict(DictBase):
    HOMEPAGE_URL = "https://terms.naer.edu.tw/"
    API = "https://terms.naer.edu.tw/search?q={word}&field=ti&match=smart"

    @property
    def provider(self):
        return "naer"

    @property
    def title(self):
        return "國家教育研究院 - 雙語詞彙、學術名詞暨辭書資訊網"

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        self.color.print("{}".format(content["title"]), "lyellow")
        print()

        for index, source in enumerate(content["sources"]):
            self.color.print("{:-2d}. ".format(index + 1), end="")
            self.color.print(source, "lred")
            for en, zhtw in content["sources"][source]:
                if len(zhtw) < 30:
                    self.color.print(zhtw, "lindigo", indent=5, end=":  ")
                    print(en)
                else:
                    self.color.print(zhtw, "lindigo", indent=5)
                    self.color.print(en, indent=5)

    def query(self, word: str):
        requests.packages.urllib3.disable_warnings()
        content = self._get_raw(word, verify=False)

        data = {"title": word, "sources": defaultdict(list)}
        soup = BeautifulSoup(content, "html.parser")
        for tr in soup.find_all("tr", {"class": "dash"}):
            source = (
                tr.find("td", attrs={"class": "sourceW"}).find("a").text
            ).strip()
            en = tr.find("td", attrs={"class": "ennameW"}).text.strip()
            zhtw = tr.find("td", attrs={"class": "zhtwnameW"}).text.strip()
            data["sources"][source].append((en, zhtw))

        if len(data["sources"]) == 0:
            raise NotFoundError(word)

        record = Record(
            word=word, content=json.dumps(data), source=self.provider
        )
        return record



from pytest import raises
from unittest.mock import Mock, patch

from zdict.dictionaries.jisho import JishoDict
from zdict.exceptions import NotFoundError
from zdict.zdict import get_args


class TestJishoDict:
    @classmethod
    def setup_class(cls):
        cls.dict = JishoDict(get_args())
        cls.word = 'bird'
        cls.source = 'jisho'
        cls.record = cls.dict.query(cls.word)

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.source
        del cls.word
        del cls.record

    def test_provider(self):
        assert self.dict.provider == 'jisho'

    def test_title(self):
        assert self.dict.title == 'Jisho'

    def test__get_url(self):
        url = (
            'https://jisho.org/api/v1/search/words?keyword={}'
        ).format(self.word)
        assert url == self.dict._get_url(self.word)

    def test_show(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.args.verbose = False
        self.dict.show(self.record)

    def test_show_verbose(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.args.verbose = True
        self.dict.show(self.record)

    @patch('zdict.dictionaries.jisho.Record')
    def test_query_normal(self, Record):
        self.dict.args.verbose = False
        self.dict.query(self.word)
        Record.assert_called_with(
            word=self.word,
            content=self.record.content,
            source=self.source,
        )

    @patch('zdict.dictionaries.jisho.Record')
    def test_query_verbose(self, Record):
        self.dict.args.verbose = True
        self.dict.query(self.word)
        Record.assert_called_with(
            word=self.word,
            content=self.record.content,
            source=self.source,
        )

    def test_query_not_found(self):
        self.dict._get_raw = Mock(return_value='{"data": []}')
        with raises(NotFoundError):
            self.dict.query(self.word)
        self.dict._get_raw.assert_called_with(self.word)



class QueryError(Exception):
    def __init__(self, word, status_code):
        self.word = word
        self.status_code = status_code

    def __str__(self):
        return '"{}" query failed on http[{}].'.format(
            self.word, self.status_code
        )


class NotFoundError(Exception):
    def __init__(self, word):
        self.word = word

    def __str__(self):
        return '"{}" not found!'.format(self.word)


class NoNetworkError(Exception):
    def __str__(self):
        return 'No Network Connection!'


class TimeoutError(Exception):
    def __str__(self):
        return 'Query timeout!'


class UnexpectedError(Exception):
    def __init__(self):
        print(
            '\n'.join((
                '',
                "Unexpected error happened.",
                "Please report the error message below to",
                "https://github.com/zdict/zdict/issues",
                '',
            ))
        )


class APIKeyError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg



import sys
import unittest

from contextlib import redirect_stdout
from io import StringIO
from unittest.mock import patch

from zdict.zdict import main


class TestMultiprocessing(unittest.TestCase):
    def test_multiprocessing(self):
        testargs = ['', '-j', '2', '-d', '-dt', 'yahoo', 'test']
        with patch.object(sys, 'argv', new=testargs):
            f1 = StringIO()
            with redirect_stdout(f1):
                main()

        testargs = ['', '-j', '-d', '-dt', 'yahoo', 'test']
        with patch.object(sys, 'argv', new=testargs):
            f2 = StringIO()
            with redirect_stdout(f2):
                main()

        testargs = ['', '-d', '-dt', 'yahoo', 'test']
        with patch.object(sys, 'argv', new=testargs):
            f3 = StringIO()
            with redirect_stdout(f3):
                main()

        result1 = f1.getvalue().strip()
        result2 = f2.getvalue().strip()
        result3 = f3.getvalue().strip()

        assert result1 == result2 == result3

    def test_multiprocessing_error(self):
        with self.assertRaises(SystemExit):
            testargs = ['', '-j', '0', '-d', '-dt', 'yahoo', 'test']
            with patch.object(sys, 'argv', new=testargs):
                f1 = StringIO()
                with redirect_stdout(f1):
                    main()

        with self.assertRaises(SystemExit):
            testargs = ['', '-j', '-1', '-d', '-dt', 'yahoo', 'test']
            with patch.object(sys, 'argv', new=testargs):
                f1 = StringIO()
                with redirect_stdout(f1):
                    main()



import re
import random
import importlib
import importlib.util

from zdict.models import Record
from zdict.utils import Color


def import_pyjokes_module():
    if importlib.util.find_spec('pyjokes'):
        return importlib.import_module('pyjokes')


def get_pyjoke(pyjokes, word: str):
    if not pyjokes:
        return

    try:
        # very basic string searching in jokes
        r = random.choice(
            list(filter(
                lambda j: word in map(
                    lambda x: ''.join(c for c in x if c.isalnum()),
                    j.split()
                ),
                pyjokes.get_jokes()
            ))
        )
    except IndexError:
        return
    else:
        return Record(word=word, content=r, source='pyjokes')


def show_pyjoke(record: Record):
    if not record:
        return

    for i, s in enumerate(
        re.split(r'\b({})\b'.format(record.word), record.content)
    ):
        Color.print(
            s,
            'lindigo' if i % 2 else 'indigo',
            end=''
        )

    print('\n\n', end='')


def lookup_pyjokes(word: str):
    pyjokes = import_pyjokes_module()

    if not pyjokes:
        return

    record = get_pyjoke(pyjokes, word)
    if record:
        Color.print('[pyjokes]', 'blue')
        show_pyjoke(record)



from pytest import raises
from unittest.mock import patch

from zdict.exceptions import NotFoundError
from zdict.zdict import get_args

from zdict.dictionaries.apc import ApcDict


class TestApcDict:
    @classmethod
    def setup_class(cls):
        cls.dict = ApcDict(get_args())

        # You may want to change words to some certain test cases.
        cls.words = ["utux", "tux"]
        cls.not_found_word = "dsafwwe"

        # Setup normal query data
        cls.dict.args.verbose = False
        cls.records = [cls.dict.query(word) for word in cls.words]

        # Setup verbose query data
        cls.dict.args.verbose = True
        cls.verbose_records = [cls.dict.query(word) for word in cls.words]

        # Change back to default verbose config
        cls.dict.args.verbose = False

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.words
        del cls.not_found_word
        del cls.records
        del cls.verbose_records

    def test_provider(self):
        assert self.dict.provider == "apc"

    def test_title(self):
        assert self.dict.title == "原住民族委員會 - 原住民族語言線上詞典"

    def test__get_url(self):
        url = "https://e-dictionary.apc.gov.tw/search/list.htm"
        assert url == self.dict._get_url("test")

    def test_show(self):
        self.dict.args.verbose = False

        for record in self.records:
            self.dict.show(record)

    def test_show_verbose(self):
        self.dict.args.verbose = True

        for record in self.verbose_records:
            self.dict.show(record)

    @patch("zdict.dictionaries.apc.Record")
    def test_query_normal(self, Record):
        self.dict.args.verbose = False

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word, content=self.records[i].content, source="apc",
            )

    @patch("zdict.dictionaries.apc.Record")
    def test_query_verbose(self, Record):
        self.dict.args.verbose = True

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word,
                content=self.verbose_records[i].content,
                source="apc",
            )

    def test_query_not_found(self):
        # Trigger NotFoundError intentionally and see if it works.
        with raises(NotFoundError):
            self.dict.query(self.not_found_word)



from pytest import raises
from unittest.mock import Mock, patch

from zdict.exceptions import NotFoundError
from zdict.dictionaries.yahoo import YahooDict
from zdict.zdict import get_args


class TestyDict:
    @classmethod
    def setup_class(cls):
        cls.dict = YahooDict(get_args())
        cls.words = ['style', 'metadata', 'apples', 'google', 'hold on']
        cls.records = [cls.dict.query(word) for word in cls.words]

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.words
        del cls.records

    def test_provider(self):
        assert self.dict.provider == 'yahoo'

    def test_title(self):
        assert self.dict.title == 'Yahoo Dictionary'

    def test__get_url(self):
        url = 'https://tw.dictionary.yahoo.com/dictionary?p=test'
        assert url == self.dict._get_url('test')

    def test_show(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.args.verbose = False

        for record in self.records:
            self.dict.show(record)

    def test_show_verbose(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.args.verbose = True

        for record in self.records:
            self.dict.show(record)

    @patch('zdict.dictionaries.yahoo.Record')
    def test_query_normal(self, Record):
        self.dict.args.verbose = False

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word,
                content=self.records[i].content,
                source='yahoo',
            )

    @patch('zdict.dictionaries.yahoo.Record')
    def test_query_verbose(self, Record):
        self.dict.args.verbose = True

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word,
                content=self.records[i].content,
                source='yahoo',
            )

    def test_query_not_found(self):
        self.dict._get_raw = Mock(return_value='{"data": []}')
        with raises(NotFoundError):
            self.dict.query(self.words[0])
        self.dict._get_raw.assert_called_with(self.words[0])



from pytest import raises
from unittest.mock import Mock, patch

from zdict.exceptions import NotFoundError
from zdict.zdict import get_args
from zdict.dictionaries.itaigi import iTaigiDict


class TestiTaigiDict:
    @classmethod
    def setup_class(cls):
        cls.dict = iTaigiDict(get_args())

        cls.words = ['芭樂', '測試']

        # Set query_timeout from 5 seconds to 60 seconds,
        # so it won't timeout that often.
        cls.dict.args.query_timeout = 60
        # Setup normal query data
        cls.dict.args.verbose = False
        cls.records = [cls.dict.query(word) for word in cls.words]

        # Setup verbose query data
        cls.dict.args.verbose = True
        cls.verbose_records = [cls.dict.query(word) for word in cls.words]

        # Change back to default verbose config
        cls.dict.args.verbose = False

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.words
        del cls.records
        del cls.verbose_records

    def test_provider(self):
        assert self.dict.provider == 'itaigi'

    def test_title(self):
        assert self.dict.title == 'iTaigi - 愛台語'

    def test__get_url(self):
        # Change url for the new dict and delete this comment
        url = 'https://itaigi.tw/平臺項目列表/揣列表?關鍵字=芭樂'
        assert url == self.dict._get_url('芭樂')

    def test_show(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.args.verbose = False

        for record in self.records:
            self.dict.show(record)

    def test_show_verbose(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.args.verbose = True

        for record in self.verbose_records:
            self.dict.show(record)

    @patch('zdict.dictionaries.itaigi.Record')
    def test_query_normal(self, Record):
        self.dict.args.verbose = False

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word,
                content=self.records[i].content,
                source='itaigi',
            )

    @patch('zdict.dictionaries.itaigi.Record')
    def test_query_verbose(self, Record):
        self.dict.args.verbose = True

        for i, word in enumerate(self.words):
            self.dict.query(word)
            Record.assert_called_with(
                word=word,
                content=self.verbose_records[i].content,
                source='itaigi',
            )

    def test_query_not_found(self):
        self.dict._get_raw = Mock(return_value='{"列表": []}')
        with raises(NotFoundError):
            self.dict.query(self.words[0])
        self.dict._get_raw.assert_called_with(self.words[0])



from zdict.loader import _is_dict, get_dictionary_map
from zdict.dictionary import DictBase

from unittest.mock import MagicMock, Mock, patch


@patch('builtins.issubclass', return_value=True)
def test__is_dict_normal(issubclass_):
    assert _is_dict(Mock) is True
    issubclass_.assert_called_with(Mock, DictBase)


@patch('builtins.issubclass', return_value=True)
def test__is_dict_DictBase(issubclass_):
    assert _is_dict(DictBase) is False
    issubclass_.assert_called_with(DictBase, DictBase)


@patch('builtins.issubclass', side_effect=TypeError)
def test__is_dict_not_class(issubclass_):
    import sys

    assert _is_dict(sys) is False
    issubclass_.called


@patch('zdict.loader.getmembers')
@patch('zdict.loader.import_module', return_value='mock')
@patch('zdict.loader.os.listdir')
def test_get_dictionary_map(listdir, import_module, getmembers):
    # prepare mocked return value
    MockDict = MagicMock(name='MockDict')
    MockDict().provider = 'mock'
    listdir.return_value = [
        '__init__.py',
        '_strange.py',
        'mock.py',
        'template.py',  # the excluded files
        'not_a_py.rst',
    ]
    getmembers.return_value = [('mock', MockDict)]

    dict_map = get_dictionary_map()

    assert dict_map == {'mock': MockDict}
    assert listdir.called
    import_module.assert_called_with('zdict.dictionaries.mock')
    assert import_module.call_count == 1
    getmembers.assert_called_with('mock', predicate=_is_dict)
    assert MockDict.called



from pytest import raises
from unittest.mock import patch

from zdict.dictionaries.urban import UrbanDict
from zdict.exceptions import NotFoundError
from zdict.zdict import get_args


class TestUrbanDict:
    @classmethod
    def setup_class(cls):
        cls.dict = UrbanDict(get_args())
        cls.not_found_word = 'some_not_existing_word'
        cls.word = 'urban'
        cls.record = cls.dict.query(cls.word)

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.not_found_word
        del cls.word
        del cls.record

    def test_provider(self):
        assert self.dict.provider == 'urban'

    def test__get_url(self):
        uri = 'http://api.urbandictionary.com/v0/define?term=mock'
        assert self.dict._get_url('mock') == uri

    def test_query_not_found(self):
        with raises(NotFoundError):
            self.dict.query(self.not_found_word)

    @patch('zdict.dictionaries.urban.Record')
    def test_query_normal(self, Record):
        self.dict.query(self.word)
        Record.assert_called_with(
            word=self.word,
            content=self.record.content,
            source='urban',
        )

    def test_show(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.show(self.record)



import json
import requests

from collections import defaultdict

from bs4 import BeautifulSoup

from zdict.dictionary import DictBase
from zdict.exceptions import NotFoundError
from zdict.models import Record


class ApcDict(DictBase):
    BASE_URL = "https://e-dictionary.apc.gov.tw"

    # To get the result of APC online dictionary
    # We have to send a POST request to the POST_API first
    # (with the word we want to query in the JSON payload)
    # Then, the webpage will save result in the cookies
    # we use the cookies to request the API with GET
    # It will shows the result.
    # Then, we parse the webpage.
    POST_API = "{}/index.aspx/_crossTribes".format(BASE_URL)
    API = "{}/search/list.htm".format(BASE_URL)

    '''
    # Ref:
    #   http://ilrdc.tw/research/athousand/area16.php
    #   原住民族語言研究中心 - 16族族語學習詞表
    NOT_FOUND_WORD_INDIGENOUS = [
        "awaay",  # 南勢/秀姑巒/海岸/馬蘭/恆春阿美語
        "ungat",  # 賽考利克/澤敖/萬大/宜蘭澤敖利泰雅語
        "ukas",  # 汶水泰雅語
        "uka",  # 四季泰雅語
        "neka",  # 東/北/中/南排灣語
        "uka",  # 卓群/卡群/丹群/巒群/郡群布農語
        "uniyan",  # 南王/知本/初鹿/建和卑南語
        "iniyan",  # 知本/建和卑南語
        "kadrwa",  # 東/霧台魯凱語
        "kadruwa",  # 大武魯凱語
        "kadrowa",  # 多納魯凱語
        "tédra",  # 茂林魯凱語
        "akaodho",  # 萬山魯凱語
        "uk'a",  # 鄒語
        "'okay",  # 賽夏語
        "'oka'",  # 賽夏語
        "'inoka'",  # 賽夏語
        "abo",  # 雅美語
        "uka",  # 邵語
        "mai",  # 噶瑪蘭語
        "ini",  # 太魯閣語
        "ungat",  # 太魯閣語
        "nai'",  # 撒奇萊雅語
        "ungac",  # 都達語 (賽德克語)
        "uka",  # 都達語 (賽德克語)
        "ini",  # 德固達雅語 (賽德克語)
        "ungac",  # 德路固語 (賽德克語)
        "ungat",  # 德路固語 (賽德克語)
        "uka'a",  # 拉阿魯哇語
        "'akia",  # 卡那卡那富語
    ]
    '''

    @property
    def provider(self):
        return "apc"

    @property
    def title(self):
        return "原住民族委員會 - 原住民族語言線上詞典"

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        self.color.print("{}".format(content["title"]), "lyellow")
        print()

        self.color.print("精確搜尋結果", "lmagenta")
        for index, source in enumerate(content["exact_sources"]):
            self.color.print("{:-2d}. ".format(index + 1), end="")
            self.color.print(source, "lred", end=": ")
            self.color.print(
                content["exact_sources"][source]["title"],
                "lindigo",
            )
            for i, d in enumerate(content["exact_sources"][source]["defs"]):
                self.color.print("{i:-2d}. {d}".format(i=i + 1, d=d), indent=4)

            self.color.print("detail: ", "indigo", end="", indent=4)
            self.color.print(content["exact_sources"][source]["link"], "white")
            print()
        if not content["exact_sources"]:
            print()
            self.color.print("無精確搜尋結果", "green")
            print()

        self.color.print("模糊搜尋結果", "lmagenta")
        for index, source in enumerate(content["fuzzy_sources"]):
            self.color.print("{:-2d}. ".format(index + 1), end="")
            self.color.print(source, "red")

            for word in content["fuzzy_sources"][source]:
                self.color.print(word["title"], "lindigo", indent=5)
                for i, d in enumerate(word["defs"]):
                    self.color.print(
                        "{i:-2d}. {d}".format(i=i + 1, d=d),
                        indent=6,
                    )
            print()
        if not content["fuzzy_sources"]:
            print()
            self.color.print("無模糊搜尋結果", "green")
            print()

    def query(self, word: str):
        r = requests.post(
            self.POST_API,
            json={
                "c": "1",
                "t": "all",
                "q": word,
            },
        )
        content = self._get_raw(word, cookies=r.cookies)

        data = {
            "title": word,
            "exact_sources": defaultdict(list),
            "fuzzy_sources": defaultdict(list),
        }
        soup = BeautifulSoup(content, "html.parser")

        # Exact matching
        exact = soup.find(id="accordion_cross")
        if exact:
            for div in exact.find_all("div", {"class": "panel"}):
                title = div.find("div", {"class": "title"}).find("a").text
                defs = [
                    div.find("strong", {"class": "word"})
                    .find_all("span")[-1]
                    .text.strip()
                ]
                source = div.find("strong", {"class": "race"}).text

                for i in div.find_all("li")[1:]:
                    d = i.find("strong", {"class": "word"})
                    if d:
                        defs.append(d.find_all("span")[-1].text.strip())

                link = (
                    self.BASE_URL
                    + div.find("a", {"class": "btn-more"})["href"]
                )
                data["exact_sources"][source] = {
                    "title": title,
                    "defs": defs,
                    "link": link,
                }

        # Fuzzy matching
        fuzzy = soup.find(id="accordion")
        if fuzzy:
            for div in fuzzy.find_all("div", {"class": "panel"}):
                title = div.find("div", {"class": "title"}).find("a").text
                defs = [
                    div.find("strong", {"class": "word"})
                    .find_all("span")[-1]
                    .text.strip()
                ]
                source = div.find("strong", {"class": "race"}).text

                for i in div.find_all("li")[1:]:
                    d = i.find("strong", {"class": "word"})
                    if d:
                        defs.append(d.find_all("span")[-1].text.strip())

                link = (
                    self.BASE_URL
                    + div.find("div", {"class": "title"}).find("a")["href"]
                )
                data["fuzzy_sources"][source].append(
                    {"title": title, "defs": defs, "link": link}
                )

        if not exact and not fuzzy:
            raise NotFoundError(word)

        record = Record(
            word=word,
            content=json.dumps(data),
            source=self.provider,
        )
        return record



from pytest import raises
from unittest.mock import patch

from zdict.dictionaries.spanish import SpanishDict
from zdict.exceptions import NotFoundError
from zdict.zdict import get_args


class TestSpanishDict:
    @classmethod
    def setup_class(cls):
        cls.dict = SpanishDict(get_args())
        cls.en_word = 'Spanish'
        cls.en_record = cls.dict.query(cls.en_word)
        cls.es_word = 'tranquilo'
        cls.es_record = cls.dict.query(cls.es_word)
        cls.not_found_word = 'asdfsdf'

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.en_word
        del cls.en_record
        del cls.es_word
        del cls.es_record
        del cls.not_found_word

    def test_provider(self):
        assert self.dict.provider == 'spanish'

    def test_title(self):
        assert self.dict.title == 'SpanishDict'

    def test__get_url(self):
        url = 'https://www.spanishdict.com/translate/{}'.format(self.en_word)
        assert url == self.dict._get_url(self.en_word)

        url = 'https://www.spanishdict.com/translate/{}'.format(self.es_word)
        assert url == self.dict._get_url(self.es_word)

    def test_show(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.show(self.en_record)
        self.dict.show(self.es_record)

    @patch('zdict.dictionaries.spanish.Record')
    def test_query_normal(self, Record):
        self.dict.query(self.en_word)
        Record.assert_called_with(
            word=self.en_word,
            content=self.en_record.content,
            source='spanish',
        )

        self.dict.query(self.es_word)
        Record.assert_called_with(
            word=self.es_word,
            content=self.es_record.content,
            source='spanish',
        )

    def test_query_not_found(self):
        with raises(NotFoundError):
            self.dict.query(self.not_found_word)
